#include "CallObserver.h"
#include "CallManager.h"
#include <resip/stack/SdpContents.hxx>
#include <resip/dum/ServerInviteSession.hxx>
#include "AppDialogSetEndCommand.h"
#include <sstream>

using namespace resip;

CallManager::CallManager(UserAgent *userAgent,RTPSessionFactory *rtpSessionFactory){
    this->mRtpSessionFactory = rtpSessionFactory;
    this->mUserAgent = userAgent;
    this->mUserAgent->getDUM()->setInviteSessionHandler(this);
}

CallManager::~CallManager(){
}

void CallManager::release(Call *call)
{
    AppDialogSetEndCommand* cmd = new AppDialogSetEndCommand(call->getHandle());
    this->mUserAgent->getDUM()->post(cmd);
}

Call* CallManager::call(std::string destination)
{
    if (!this->mUserAgent->isRegistered()) return 0;

    DialogUsageManager* dum = this->mUserAgent->getDUM();
    AccountSettings settings = this->mUserAgent->getAccountSettings();

    NameAddr dest;
    std::string uri = "sip:" + destination + "@" + settings.mProxy;
    dest = NameAddr(Uri(uri.c_str()));

    SharedPtr<UserProfile> up(dum->getMasterUserProfile());
    RTPSession *rtpSession = this->mRtpSessionFactory->createRTPSession();
    Call *call = new Call(*dum);
    call->setRTPSession(rtpSession);

    std::stringstream ss;
    ss<<"v=0\r\n"
        "o=- 0 0 IN IP4 "<<rtpSession->getLocalIP()<<"\r\n"
        "s="<<this->mUserAgent->getName()<<"\r\n"
        "c=IN IP4 "<<rtpSession->getLocalIP()<<"\r\n"
        "t=0 0\r\n"
        "m=audio "<<rtpSession->getLocalPort()<<" RTP/AVP 0 \r\n" // only support G.711 uLaw
        "a=rtpmap:0 pcmu/8000\r\n";
    Data txt(ss.str().c_str());
    HeaderFieldValue hfv(txt.data(), txt.size());
    Mime type("application", "sdp");
    SdpContents sdp(&hfv, type);
    UInt64 currentTime = Timer::getTimeMicroSec();
    sdp.session().origin().getSessionId() = currentTime;
    sdp.session().origin().getVersion() = currentTime;

    SharedPtr<SipMessage> msg = dum->makeInviteSession(dest, up, &sdp, call);

    dum->send(msg);
    return call;
}

void CallManager::addObserver(CallObserver *obs)
{
    this->mCallObservers.push_back(obs);
}


void CallManager::CallManager::onSuccess(resip::ClientRegistrationHandle h, const resip::SipMessage& response)
{
}

void CallManager::onFailure(resip::ClientRegistrationHandle, const resip::SipMessage& msg)
{
}

void CallManager::onMessage(resip::Handle<resip::InviteSession>, const resip::SipMessage& msg)
{
}

void CallManager::onMessageSuccess(resip::Handle<resip::InviteSession>, const resip::SipMessage&)
{
}

void CallManager::onMessageFailure(resip::Handle<resip::InviteSession>, const resip::SipMessage&)
{
}

void CallManager::onFailure(resip::ClientInviteSessionHandle cis, const resip::SipMessage& msg)
{
    //TODO
}

void CallManager::onForkDestroyed(resip::ClientInviteSessionHandle)
{
}

void CallManager::onInfoSuccess(resip::InviteSessionHandle, const resip::SipMessage& msg)
{
}

void CallManager::onInfoFailure(resip::InviteSessionHandle, const resip::SipMessage& msg)
{
}

void CallManager::onProvisional(resip::ClientInviteSessionHandle cis, const resip::SipMessage& msg)
{
    Call *call = (Call*)cis->getAppDialogSet().get();
//  call->clientInviteSessionHandle = cis;

    /*std::stringstream sst;
    sst << msg.brief();
    printf("%s\r\n\r\n",sst.str().c_str());*/

    switch (msg.header(h_StatusLine).statusCode())
    {
    case 100:
        
       // call->onTrying();
        break;
    case 180:
        //call->onRinging();
        break;
    }

}

void CallManager::onConnected(resip::ClientInviteSessionHandle cis, const resip::SipMessage& msg)
{
}

void CallManager::onConnected(resip::InviteSessionHandle is, const resip::SipMessage& msg)
{
    printf("CallManager::onConnected\r\n");
    Call *call = (Call*)is->getAppDialogSet().get();
    if (call)
    {
        RTPSession *rtpSession = call->getRTPSession();
        rtpSession->start();
        for (std::list<CallObserver*>::iterator it = this->mCallObservers.begin();it!=this->mCallObservers.end();it++)
        {
            (*it)->onIncommingCallAnswered(call);        
        }
    }
}

void CallManager::onStaleCallTimeout(resip::ClientInviteSessionHandle)
{
}

void CallManager::onRedirected(resip::ClientInviteSessionHandle, const resip::SipMessage& msg)
{
}

void CallManager::onAnswer(resip::InviteSessionHandle is, const resip::SipMessage& msg, const resip::SdpContents& sdp)
{
    Call *call = (Call*)is->getAppDialogSet().get();

    const char* peerIP = sdp.session().origin().getAddress().c_str();
    unsigned int peerPort = sdp.session().media().front().port();

    RTPSession *rtpSession = call->getRTPSession();
    rtpSession->setPeerAddress(peerIP, peerPort);
    rtpSession->start();

    for (std::list<CallObserver*>::iterator it = this->mCallObservers.begin();it!=this->mCallObservers.end();it++)
    {
        (*it)->onCallAnswered(call);
    }

}

void CallManager::onEarlyMedia(resip::ClientInviteSessionHandle, const resip::SipMessage& msg, const resip::SdpContents& sdp)
{
}

void CallManager::onOfferRequired(resip::InviteSessionHandle, const resip::SipMessage& msg)
{
}

void CallManager::onOfferRejected(resip::Handle<resip::InviteSession>, const resip::SipMessage *msg)
{
}

void CallManager::onInfo(resip::InviteSessionHandle is, const resip::SipMessage& msg)
{
    Call *call = dynamic_cast<Call*>(is->getAppDialogSet().get());

    Mime dtmf(resip::Data("application"), resip::Data("dtmf-relay"));
    Contents* contents = msg.getContents();
    Mime mime = msg.getContents()->getType();
    if (mime.type() == "application" && mime.subType() == "dtmf-relay")
    {
        Data data = msg.getContents()->getBodyData();
        std::istringstream iss(data.c_str());
        std::string digit;
        std::getline(iss, digit,'=');
        iss >> digit;

        call->mDigitQueue += digit;
        for (std::list<CallObserver*>::iterator it = this->mCallObservers.begin();it!=this->mCallObservers.end();it++)
        {
            (*it)->onDTMF(call,digit);
        }

        is->acceptNIT();
    }
}

void CallManager::onRefer(resip::InviteSessionHandle, resip::ServerSubscriptionHandle, const resip::SipMessage& msg)
{
}

void CallManager::onReferAccepted(resip::InviteSessionHandle, resip::ClientSubscriptionHandle, const resip::SipMessage& msg)
{
}

void CallManager::onReferRejected(resip::InviteSessionHandle, const resip::SipMessage& msg)
{
}

void CallManager::onReferNoSub(resip::Handle<resip::InviteSession>, const resip::SipMessage&)
{
}

void CallManager::onRemoved(resip::ClientRegistrationHandle)
{
}

int CallManager::onRequestRetry(resip::ClientRegistrationHandle, int retrySeconds, const resip::SipMessage& response)
{
}

void CallManager::onNewSession(resip::ServerInviteSessionHandle sis, resip::InviteSession::OfferAnswerType oat, const resip::SipMessage& msg)
{
    
    //TODO: this is an "iriginated" event
    Call *call = dynamic_cast<Call*>(sis->getAppDialogSet().get());
    if (call)
    {
        call->setIncomming(true);
  //      printf("CallManager::onOffer\r\n");
    }

}

void CallManager::onNewSession(resip::ClientInviteSessionHandle cis, resip::InviteSession::OfferAnswerType oat, const resip::SipMessage& msg)
{
    //TODO
    Call *call = (Call*)cis->getAppDialogSet().get();

}

void CallManager::onTerminated(resip::InviteSessionHandle is, resip::InviteSessionHandler::TerminatedReason reason, const resip::SipMessage* msg)
{


//printf("onTerminated: reason=%i\r\n",reason);
    if (reason==LocalBye)
    {
//        is->dispatchBye(*msg);
    }

    Call *call = (Call*)is->getAppDialogSet().get();
    if (call)
    {
        RTPSession *rtp = call->getRTPSession();
        if (rtp)
        {
            rtp->stop();
        }
        for (std::list<CallObserver*>::iterator it = this->mCallObservers.begin();it!=this->mCallObservers.end();it++)
        {
            (*it)->onCallTerminated(call);
        }
    }

}


void CallManager::onOffer(resip::InviteSessionHandle is, const resip::SipMessage& msg, const resip::SdpContents& sdp)
{
    Call *call = dynamic_cast<Call*>(is->getAppDialogSet().get());
    if (call)
    {

        RTPSession *rtpSession = this->mRtpSessionFactory->createRTPSession();
        call->setRTPSession(rtpSession);

        const char* peerIP = sdp.session().origin().getAddress().c_str();
        unsigned int peerPort = sdp.session().media().front().port();
        rtpSession->setPeerAddress(peerIP, peerPort);
        call->mServerInviteSession = dynamic_cast<ServerInviteSession*>(is.get());

        for (std::list<CallObserver*>::iterator it = this->mCallObservers.begin();it!=this->mCallObservers.end();it++)
        {
            (*it)->onIncommingCall(call);
        }
    }
}

void  CallManager::answerIncomming(Call* call)
{

        DialogUsageManager* dum = this->mUserAgent->getDUM();
        AccountSettings settings = this->mUserAgent->getAccountSettings();
        RTPSession *rtpSession = call->getRTPSession();

        std::stringstream ss;
        ss<<"v=0\r\n"
            "o=- 0 0 IN IP4 "<<rtpSession->getLocalIP()<<"\r\n"
            "s="<<this->mUserAgent->getName()<<"\r\n"
            "c=IN IP4 "<<rtpSession->getLocalIP()<<"\r\n"
            "t=0 0\r\n"
            "m=audio "<<rtpSession->getLocalPort()<<" RTP/AVP 0 \r\n" // only support G.711 uLaw
            "a=rtpmap:0 pcmu/8000\r\n";
        Data txt(ss.str().c_str());
        HeaderFieldValue hfv(txt.data(), txt.size());
        Mime type("application", "sdp");
        SdpContents sdp2(&hfv, type);
        UInt64 currentTime = Timer::getTimeMicroSec();
        sdp2.session().origin().getSessionId() = currentTime;
        sdp2.session().origin().getVersion() = currentTime;

        if (call->mServerInviteSession)
        {
            call->mServerInviteSession->provideAnswer(sdp2);
            printf("Answering incomming call\r\n");
            call->mServerInviteSession->accept();

        }
}

