#include "RTPSession.h"
#include "SoundFile.h"
#include "SilenceSound.h"
#include "config.h"
#include <sstream>

void *threadStarter(void *p)
{
    RTPSession *f = (RTPSession*)p;
    f->run();
}


RTPSession::RTPSession()
{
    this->mSession=rtp_session_new(RTP_SESSION_SENDRECV);
    this->mLocalIP = "";
    this->mLocalPort=0;
    pthread_mutex_init(&this->mQueueLock,0);
}

RTPSession::~RTPSession(){
    this->stop();

    // Clear the queue
    while(this->mSoundQueue.size())
    {
        ISound *s = this->mSoundQueue.front();
        delete s;
        this->mSoundQueue.pop();
    }

    pthread_mutex_destroy(&this->mQueueLock);

}


void RTPSession::start()
{
    this->mAbort = false;

    rtp_session_set_local_addr(mSession,this->mLocalIP.c_str(),this->mLocalPort);
    rtp_session_set_scheduling_mode(mSession,1);
    rtp_session_set_blocking_mode(mSession,1);
    rtp_session_set_connected_mode(mSession,TRUE);
    rtp_session_set_remote_addr(mSession,this->mPeerIP.c_str(),this->mPeerPort);
    rtp_session_set_payload_type(mSession,0);

    //TODO: use thread pool
    pthread_create(&this->mThreadHandle, 0, threadStarter, (void*)this);

}

void RTPSession::stop()
{

    if (!this->mAbort && this->mSession!=0)
    {
        this->mAbort = true;
        printf("Stopping RTP session\r\n");
        int ret = pthread_join(this->mThreadHandle,0);
        printf("RTP session stopped(join=%i)\r\n",ret);
        rtp_session_destroy(this->mSession);
        this->mSession = 0;
    }
}

void RTPSession::silence(int seconds)
{
    SilenceSound *sound = new SilenceSound(seconds);
    pthread_mutex_lock(&this->mQueueLock);
    this->mSoundQueue.push(sound);
    pthread_mutex_unlock(&this->mQueueLock);
}

void RTPSession::play(const char *filename)
{
    SoundFile *file = new SoundFile();
    if (file->open(filename))
    {
        pthread_mutex_lock(&this->mQueueLock);
        this->mSoundQueue.push(file);        
        pthread_mutex_unlock(&this->mQueueLock);
    } else {
        delete file;
    }
}


//WARNING: currently only support 0-9
void RTPSession::playNumber(int number)
{
    char c[6];
    c[0] = number+48;
    c[1] = '.';
    c[2] = 'w';
    c[3] = 'a';
    c[4] = 'v';
    c[5] = 0;
    std::stringstream ss;
    ss << SOUND_PATH << (char*)&c;
    this->play(ss.str().c_str());

    
}

void RTPSession::setPeerAddress(std::string ip, unsigned int peerPort)
{
    this->mPeerIP = ip;
    this->mPeerPort= peerPort;
}

void RTPSession::setLocalAddress(std::string ip, unsigned int port)
{
    this->mLocalPort = port;
    this->mLocalIP = ip;
}

unsigned int RTPSession::getLocalPort()
{
    return this->mLocalPort;
}

std::string RTPSession::getLocalIP()
{
    return this->mLocalIP;
}


void RTPSession::clearQueue()
{
    pthread_mutex_lock(&this->mQueueLock);
    this->mSoundQueue=std::queue<ISound*>();
    this->mInterruptCurrentSound = true;
    pthread_mutex_unlock(&this->mQueueLock);
}

void RTPSession::run()
{
    unsigned int user_ts;
    char buffer[CONF_RTP_PAYLOAD_SIZE];

    ISound *currentSound=0;
    bool queueWasEmpty=true;
    this->mInterruptCurrentSound=false;
    while (!this->mAbort)
    {
        if (!currentSound)
        {
            pthread_mutex_lock(&this->mQueueLock);
            if (this->mSoundQueue.size())
            {
                currentSound = this->mSoundQueue.front();
                this->mSoundQueue.pop();
                pthread_mutex_unlock(&this->mQueueLock);
                printf("Playing %s\r\n",currentSound->toString().c_str());
                queueWasEmpty=false;
                this->mInterruptCurrentSound=false;
                user_ts = rtp_session_get_current_send_ts(this->mSession);
            } else {
                pthread_mutex_unlock(&this->mQueueLock);
                if (!queueWasEmpty)
                {
                    this->notifyQueueEmpty();
                    queueWasEmpty=true;
                }
            }
            //TODO: should use wait condition when soundqueue is empty
        }


        int i = 0;
        if (currentSound)
        {
            i=currentSound->getSample(CONF_RTP_PAYLOAD_SIZE,(char*)&buffer);

            for (int n=i;n<CONF_RTP_PAYLOAD_SIZE;n++) buffer[n]=0; // pad with silence if it is too short.
            rtp_session_send_with_ts(this->mSession,(uint8_t*)&buffer,CONF_RTP_PAYLOAD_SIZE,user_ts);
            user_ts+=CONF_RTP_PAYLOAD_SIZE;

            if (i<CONF_RTP_PAYLOAD_SIZE || this->mInterruptCurrentSound)
            {
                delete currentSound;
                currentSound = 0;
            }
        }
    }

}

void RTPSession::addObserver(IRTPSessionObserver* obs)
{
    this->mObservers.push_back(obs);
}

int RTPSession::getSoundQueueSize()
{
    int size = 0;
    pthread_mutex_lock(&this->mQueueLock);
    size = this->mSoundQueue.size();
    pthread_mutex_unlock(&this->mQueueLock);
    return size;
}

void RTPSession::notifyQueueEmpty()
{
    for (std::list<IRTPSessionObserver*>::iterator it = this->mObservers.begin();it!=this->mObservers.end();it++)
    {
        (*it)->onSoundQueueEmpty();
    }
}
